@@ -1,7 +1,8 @@
 from __future__ import absolute_import
 
 import argparse
-from flask.ext.script import Command
+from flask_script import Command
+
 
 class TaskTigerCommand(Command):
     capture_all_args = True
@@ -14,10 +15,10 @@ class TaskTigerCommand(Command):
     def create_parser(self, *args, **kwargs):
         # Override the default parser so we can pass all arguments to the
         # TaskTiger parser.
-        func_stack = kwargs.pop('func_stack',())
+        func_stack = kwargs.pop('func_stack', ())
         parent = kwargs.pop('parent', None)
         parser = argparse.ArgumentParser(*args, add_help=False, **kwargs)
-        parser.set_defaults(func_stack=func_stack+(self,))
+        parser.set_defaults(func_stack=func_stack + (self,))
         self.parser = parser
         self.parent = parent
         return parser
@@ -29,5 +30,8 @@ class TaskTigerCommand(Command):
         """
 
     def run(self, args):
+        # Allow passing a callable that returns the TaskTiger instance.
+        if callable(self.tiger):
+            self.tiger = self.tiger()
         self.setup()
         self.tiger.run_worker_with_args(args)
